///Error caused by registered name
class RegisteredNameError extends Error
{
    final String _message;
    RegisteredNameError(this._message);
    @override
    String toString() => 'error: $_message';
}

///Error caused by unregistered type name
class UnregisteredNameError extends Error
{
    final String _message;
    UnregisteredNameError(this._message);
    @override
    String toString() => 'error: $_message';
}

///Ioc Container Interface
abstract class IContainer
{
    void set(Symbol name, Factory factory);

    dynamic get(Symbol name, {bool singleton = true});

    void unset(Symbol name);

    bool has(Symbol name);
    bool hasCache(Symbol name);
}

typedef Factory = dynamic Function(IContainer container);

///Ioc Container
class Container implements IContainer
{
    final Map<Symbol, Factory> _factories = {};
    final Map<Symbol, dynamic> _cached    = {};

    @override
    void set(Symbol name, Factory factory)
    {
        if(has(name)) {
            throw RegisteredNameError("'$name' has been registered");
        }

        _factories[name] = factory;
    }

    @override
    dynamic get(Symbol name, {bool singleton = true})
    {
        if(!has(name)) {
            throw UnregisteredNameError("Unregistered '$name'");
        }

        if(!singleton || !hasCache(name)) {
            return _cached[name] = _factories[name](this);
        }

        return _cached[name];
    }

    @override
    void unset(Symbol name)
    {
        _factories?.remove(name);
        _cached?.remove(name);
    }

    @override
    bool has(Symbol name)
    {
        return _factories.containsKey(name);
    }

    @override
    bool hasCache(Symbol name)
    {
        return _cached.containsKey(name);
    }
}